# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 10:11:24 2024

@author: Andrea
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import c,m_e,e,mu_0,epsilon_0
from scipy.integrate import solve_ivp
import copy
from .rad_utils import rad_parall,rad_no_parall
from .segment_utils import segment
from .fields_utils import field_dict,spatial_field_ramp

npn = np.newaxis

class beamline():
    def __init__(self):
        self.segments = {}              # empty dictionary to add beamline devices and drifts; devices will be saverd by names
        self.add_segment("drift",       # first not executed drift to obtain field interface with first real segment
                         "start_drift",
                         l=1)
        self.color_dict = {"drift":         "gray",
                           "undulator":     "yellow",
                           "ion channel":   "green",
                           "CBM":           "blue",
                           "ABP":           "red"}
    
    def add_beam(self,beam,istart=0):
        """
        Function add a beam to the beamline in a given segment position. 
        
        """
        self.istart = istart+1  # index of the starting beamline item
        self.beam = beam
        
    def add_segment(self,device,name,**kwargs):
        """
        Function to extend beamline adding segments. 
        Kwargs should include segment specs and length, 
        and will be updated with additional and derived vsariables.
        Then they will be stored in a beamline dict, together with segment name and type.
        
        """
        
        """create segment object and update segments dictionary"""
        seg = segment(device,name,**kwargs)
        d = {name:seg}
        self.segments.update(d)
        
        """update segments name list"""
        self.slist = list(self.segments.keys())
    
    def link_fields(self,seg1,seg2,z0):
        """
        Gives spatial connection between two field types.
        
        dname1,dname2: strings, names of the two devices to link
        """
        if seg1 is not None:
            """left and right bend use the same magnetic field (ref sys swapping): 
                going from left to right bend should simulate the real field change"""
            swapCBMfield = 1
            if seg1.kind=="CBM":
                if seg2.kind=="CBM" and seg1.kwargs["lr"]!=seg2.kwargs["lr"]:
                    swapCBMfield = -1
                if seg2.kind!="CBM" and seg1.kwargs["lr"]=="L":
                    swapCBMfield = -1
            
            """add offset for correct undulator phase injection"""
            z01 = 0
            z02 = 0
            if seg1.kind=="undulator":
                z01 = z0
            if seg2.kind=="undulator":
                z02 = z0
                
            """retrieve separate fields"""
            field1 = field_dict[seg1.kind]
            field2 = field_dict[seg2.kind]
            
            """define combined fields function"""
            def linked_field(x,y,z):
                """fading field 1"""
                f1 = field1(x,y,z - z01,**seg1.kwargs)*spatial_field_ramp(z,z0,updown=-1)*swapCBMfield
                """growing field 2"""
                f2 = field2(x,y,z - z02,**seg2.kwargs)*spatial_field_ramp(z,z0,updown=1)
                return f1 + f2
        
        else:
            z02 = 0
            if seg2.kind=="undulator":
                z02 = z0
            field2 = field_dict[seg2.kind]
            def linked_field(x,y,z):
                f2 = field2(x,y,z - z02,**seg2.kwargs)
                return f2
        
        return linked_field
    
    def match(self):
        """overwrites original transverse rms sizes and matches the beam over given emittances"""
        
    def LLforce(self,bx,by,bz,Ex,Ey,Ez,Bx,By,Bz):
        """
        Landau-Lifshitz radiation damping: 
        Landau, L. D. (Ed.). (2013). The classical theory of fields (Vol. 2). Elsevier. Section 76
        """
        # """CGS constants"""
        # c_CGS = 2.99792458e8*1e2  #cm/s
        # me_CGS = 9.109e-31*1e3    #g
        # e_CGS = 4.8032e-10        #statC
        # """beta-collinear unit vector (new z axis)"""
        # b = np.sqrt(bx**2 + by**2 + bz**2)
        # g = 1/np.sqrt(1-b**2)
        # nz = np.stack([bx/b,by/b,bz/b],axis=1)
        # """retrieve cylindrical coordinates"""
        # theta = np.arccos(bz/b)
        # phi = np.arctan2(bx,by)
        # """calculate new z axis"""
        # nyz = -np.sin(theta)
        # nyy = np.cos(theta)*np.cos(phi)
        # nyx = np.cos(theta)*np.sin(phi)
        # ny = np.stack([nyx,nyy,nyz],axis=1)
        # """calculate new y axis"""
        # nx = np.cross(ny,nz,axis=1)
        # """create array of basis change matrix"""
        # M = np.stack([nx,ny,nz],axis=1)
        # """basis change for E and B field (turned into CGS units)"""
        # E = np.stack([Ex,Ey,Ez],axis=1)*1e-4/2.9979
        # B = np.stack([Bx,By,Bz],axis=1)*1e4
        # En = np.sum(M*E[:,npn,:],axis=2)
        # Bn = np.sum(M*B[:,npn,:],axis=2)
        # """force calculation"""
        # field_coef = (En[:,0] - Bn[:,1])**2 + (En[:,1] + Bn[:,0])**2
        # LLF = -2*e_CGS**4*g**2/3/me_CGS**2/c_CGS**4*field_coef
        # """convert back in SI"""
        # LLF = LLF*1e-5
        
        # print()
        # print(LLF)
        
        """CGS constants"""
        c_CGS = 2.99792458e8*1e2  #cm/s
        me_CGS = 9.109e-31*1e3    #g
        e_CGS = 4.8032e-10        #statC
        """beta vector"""
        bv = np.stack([bx,by,bz],axis=1)
        b = np.sqrt(bx**2 + by**2 + bz**2)
        g = 1/np.sqrt(1-b**2)
        """fields (CGS units)"""
        E = np.stack([Ex,Ey,Ez],axis=1)*1e-4/2.9979
        B = np.stack([Bx,By,Bz],axis=1)*1e4
        """force calculation"""
        cf1 = np.cross(bv,B,axis=1)
        cf2 = np.sum(E*bv,axis=1)**2
        cf3 = E + cf1
        cf4 = np.sum(cf3*cf3,axis=1)
        field_coeff = cf4 - cf2
        LLF = -2*e_CGS**4*g**2/3/me_CGS**2/c_CGS**4*field_coeff
        """convert back in SI"""
        LLF = LLF*1e-5

        # print()
        # print(LLF)
        
        return LLF*bx/b,LLF*by/b,LLF*bz/b

    def IVP(self,t,y,npart):   
        
        """reshape variables array"""
        U = y.reshape(npart, 6)    
        x = U[:, 0]
        y = U[:, 1]
        z = U[:, 2]
        bx = U[:, 3]
        by = U[:, 4]
        bz = U[:, 5]
        g = 1/np.sqrt(1-(bx**2 + by**2 + bz**2))

        # if np.amin(z)<0:
        #     print(np.amin(z))
        #     print(np.amax(z))
        #     print()

        """calculate fields"""
        Ex,Ey,Ez,Bx,By,Bz = self.field(x,y,z)
        
        """tempoal smoothing"""
        #smooth = self.temp_field_ramp(t,self.ti,self.te)
        
        """calculate force"""
        # LFx = -e*(Ex + c*(by*Bz - bz*By))                           # Lorentz
        # LFy = -e*(Ey + c*(bz*Bx - bx*Bz))
        # LFz = -e*(Ez + c*(bx*By - by*Bx))
        # LLFx,LLFy,LLFz = self.LLforce(bx,by,bz,Ex,Ey,Ez,Bx,By,Bz)   # Landau-Lifshitz radiation damping
        # Fx = LFx + LLFx
        # Fy = LFy + LLFy
        # Fz = LFz + LLFz
        # bdF = bx*Fx + by*Fy + bz*Fz
        # # print(LLFx/LFx,LLFx,LFx)
        # # print(LLFy/LFy,LLFy,LFy)
        # # print(LLFz/LFz,LLFz,LFz)
        # # print()
        # # print()
        # # print("lorentz",np.sqrt(LFx**2 + LFy**2 + LFz**2))
        # # print("LL",np.sqrt(LLFx**2 + LLFy**2 + LLFz**2))
        
        Fx = -e*(Ex + c*(by*Bz - bz*By))                           # Lorentz
        Fy = -e*(Ey + c*(bz*Bx - bx*Bz))
        Fz = -e*(Ez + c*(bx*By - by*Bx))
        bdF = bx*Fx + by*Fy + bz*Fz

        
        """calculate acceleration"""
        ax = 1/m_e/g * (Fx - bdF*bx)
        ay = 1/m_e/g * (Fy - bdF*by)
        az = 1/m_e/g * (Fz - bdF*bz)
        
        """derivatives"""
        dxdt = bx*c
        dydt = by*c
        dzdt = bz*c
        dbxdt = ax/c
        dbydt = ay/c
        dbzdt = az/c
        
        """reshape and flatten variables array"""
        out = np.stack([dxdt, dydt, dzdt, dbxdt, dbydt, dbzdt]).transpose().flatten()
        return out
    
    def dyn(self,u0,te,teval,dtmax,seg_curr,event):
        """
        Solves inital value problem for beam dynamics
        """
        def threshold(t,y,npart):
            """
            Event to stop fine Runge-Kutta beam dynamics integration once the beam
            has fully passed from previous to current segment
            """
            U = y.reshape(npart, 6)   
            z = U[:, 2]
            return np.amin(z)
        threshold.terminal = True
        threshold.direction = 1
        u0 = np.stack(u0).transpose()
        u0 = u0.flatten()
        if event==False:
            u = solve_ivp(self.IVP, 
                          [0, te], 
                          u0, 
                          method='RK45',
                          t_eval=teval, 
                          max_step=dtmax, 
                          args=(self.beam.npart,))
        else:
            u = solve_ivp(self.IVP, 
                          [0, te], 
                          u0, 
                          method='RK45',
                          t_eval=teval, 
                          max_step=dtmax, 
                          args=(self.beam.npart,),
                          events=threshold)
        x = u.y[slice(0, 0+6*self.beam.npart, 6)].transpose()
        y = u.y[slice(1, 0+6*self.beam.npart, 6)].transpose()
        z = u.y[slice(2, 0+6*self.beam.npart, 6)].transpose()
        bx = u.y[slice(3, 0+6*self.beam.npart, 6)].transpose()
        by = u.y[slice(4, 0+6*self.beam.npart, 6)].transpose()
        bz = u.y[slice(5, 0+6*self.beam.npart, 6)].transpose()
        return x,y,z,bx,by,bz
    
    def run(self,dt,dtmax):
        """ 
        dt: output sampling time [s]
        dtmax: maximum allowed timestep for Runge-Kutta integration
            
        self.cff is used to effectively change the direction of beam bending, 
        making it easy to rectify trajectories in order to prepare beam for 
        next segment: for a left bend, first x particles coordinates and speeds 
        are flipped, then the beam is propagated "backwards" and finally x 
        coordinates and speeds are flipped back
        """
        
        """check existence of at least one segment"""
        assert len(self.segments)>0
        
        """preliminary loop over segments for empty arrays initialization"""
        self.rangelist = []                             # timestep ranges list for varisables postprocessing
        for sname in self.slist[self.istart:]:
            """extract segment"""
            seg = self.segments[sname]
            """set segment timesteps number and their value"""
            seg.add_timesteps(dt)
            """once the segment has a simulation timestep, empty variables arrays are initialized"""
            seg.init_arrays(self.beam.npart)
        
        """init integration variables"""
        x0 = copy.deepcopy(self.beam.pos0[:,0])
        y0 = copy.deepcopy(self.beam.pos0[:,1])
        z0 = copy.deepcopy(self.beam.pos0[:,2])
        bx0 = copy.deepcopy(self.beam.bet0[:,0])
        by0 = copy.deepcopy(self.beam.bet0[:,1])
        bz0 = copy.deepcopy(self.beam.bet0[:,2])
        
        """counter for timesteps"""
        nptcount = 0
        for i,sname in enumerate(self.slist[self.istart:]):
            """previous and current segment specs"""
            if i == 0:
                """always choose 'start drift' as first beamline element"""
                seg_prev = self.segments[self.slist[i]]
                i += self.istart-1
            else:
                i += self.istart-1
                seg_prev = self.segments[self.slist[i]]
            seg_curr = self.segments[self.slist[i+1]]

            """extract current segment specs"""
            npt = seg_curr.npts
            
            """extract segments names"""
            s0 = self.slist[i]
            s1 = self.slist[i+1]
            
            """push beam below z=0 line and set longitudinal position where to 
            join fields: slightly forward first beam particle"""
            z0 -= np.amax(z0)
            z_join = 0
            
            """change reference system in case of bending direction change
            and add r_bend offset (that is automatically zero for straigth elements)"""
            if seg_curr.kwargs["cff"]!=seg_prev.kwargs["cff"]:
                x0 *= -1
                bx0 *= -1
            x0 += seg_curr.kwargs["r_bend"]
            
            if self.beam.npart>1:
                """set segment field joining previous and current segment fields:
                   junction is made at first beam particle position"""
                self.field = self.link_fields(seg_prev,seg_curr,z_join)
                print()
                print(s0,"->",s1)
                
                """solve IVP along previous and current segment junction with small
                enough timestep"""
                stdz = np.std(z0)
                te_j = 10*stdz/c
                teval_j = np.linspace(0, te_j, npt, endpoint=True)
                dtmax_j = stdz/c/100
                print()
                print("dt_junc:",dtmax_j)
                print("dt_seg:",dtmax)
                print()
                u0 = (x0, y0, z0, bx0, by0, bz0)
                x,y,z,bx,by,bz = self.dyn(u0,te_j,teval_j,dtmax_j,seg_curr,event=True)
                x0 = x[-1]
                y0 = y[-1]
                z0 = z[-1]
                bx0 = bx[-1]
                by0 = by[-1]
                bz0 = bz[-1]
                
                b = np.sqrt(bx[0]**2 + by[0]**2 + bz[0]**2)
                g = 1/np.sqrt(1-b**2)
                b0 = np.sqrt(bx0**2 + by0**2 + bz0**2)
                g0 = 1/np.sqrt(1-b0**2)
            
            """set current segment field"""
            self.field = self.link_fields(None,seg_curr,z_join)
            print()
            print(s1)
            
            """solve IVP in current segment"""
            te = seg_curr.t_end
            teval = np.linspace(0, te, npt, endpoint=True)
            u0 = (x0, y0, z0, bx0, by0, bz0)
            x,y,z,bx,by,bz = self.dyn(u0,te,teval,dtmax,seg_curr,event=False)
            
            """define segment index range for array update"""
            rg = range(nptcount,nptcount+npt)
            self.rangelist.append(rg)
            
            """extend dynamic variables arrays: not yet rectified, so that full information is available for radiation evaluation"""
            seg_curr.pos[:,:,0] = x
            seg_curr.pos[:,:,1] = y
            seg_curr.pos[:,:,2] = z
            seg_curr.bet[:,:,0] = bx
            seg_curr.bet[:,:,1] = by
            seg_curr.bet[:,:,2] = bz
            
            """extend fields seen from particles during evolution: not rectified"""
            fields = self.field(x,y,z)
            seg_curr.E_field[:,:,0] = fields[0]
            seg_curr.E_field[:,:,1] = fields[1]
            seg_curr.E_field[:,:,2] = fields[2]
            seg_curr.B_field[:,:,0] = fields[3]
            seg_curr.B_field[:,:,1] = fields[4]
            seg_curr.B_field[:,:,2] = fields[5]
            
            """rectify variables for bending elements"""
            if seg_curr.kind == "CBM" or seg_curr.kind == "ABP":
                xr,yr,zr,bxr,byr,bzr = self.rectify(x,y,z,bx,by,bz,seg_curr.kwargs["r_bend"],0)
                """rms variables: after rectifying, quantities relative to reference trajectory"""
                sx = np.std(xr,axis=1)
                sy = np.std(yr,axis=1)
                sz = np.std(zr,axis=1)
                mx = np.average(xr,axis=1)
                my = np.average(yr,axis=1)
                mz = np.average(zr,axis=1)
            else:
                sx = np.std(x,axis=1)
                sy = np.std(y,axis=1)
                sz = np.std(z,axis=1)
                mx = np.average(x,axis=1)
                my = np.average(y,axis=1)
                mz = np.average(z,axis=1)
            
            """rotate last particle positions/velocity in case of bending elements"""
            if seg_curr.kind == "CBM" or seg_curr.kind == "ABP":
                r = np.sqrt(x**2 + z**2)
                be = np.sqrt(bx**2 + bz**2)
                thp = np.arctan2(z, x)
                thbe = np.arctan2(bx, bz)
                """particle head position"""
                thmax = np.amax(thp)
                """rotate angles"""
                thp -= thmax
                thbe += thmax
                """update positions and velocities with rotated angles"""
                x = r*np.cos(thp) - seg_curr.kwargs["r_bend"]
                z = r*np.sin(thp)
                bx = be*np.sin(thbe)
                bz = be*np.cos(thbe)
                print("amax z",np.amax(z[-1]))

            """extend rms variables arrays"""
            seg_curr.sigpos[:,0] = sx
            seg_curr.sigpos[:,1] = sy
            seg_curr.sigpos[:,2] = sz
            seg_curr.mupos[:,0] = mx
            seg_curr.mupos[:,1] = my
            seg_curr.mupos[:,2] = mz
            seg_curr.gam[:,:] = 1/np.sqrt(1-np.linalg.norm(seg_curr.bet,axis=2)**2)
            # plt.figure()
            # plt.plot(teval,1/np.sqrt(1-np.linalg.norm(seg_curr.bet,axis=2)**2))
            """update initial integration variables: they need to be the 
            rectified ones since each segment has its reference system. beam
            longitudinal position is initialized before next segment's z=0 line"""
            x0 = x[-1]
            y0 = y[-1]
            z0 = z[-1]-np.amax(z[-1])
            bx0 = bx[-1]
            by0 = by[-1]
            bz0 = bz[-1]
            
            """update counter"""
            nptcount += npt
    
    def rectify(self,x,y,z,bx,by,bz,r_bend,z0):
        """Rectifies coordinates for bending elements. Axis origin referred to bending radius"""
        r = np.sqrt(x**2 + z**2)
        be = np.sqrt(bx**2 + bz**2)
        thp = np.arctan2(z, x)
        thet = np.average(thp,axis=1)
        thbe = np.arctan2(bx, bz)
        xth = r - r_bend
        zth = r_bend*thp
        bexth = be*np.sin(thbe+thp)
        bezth = be*np.cos(thbe+thp)
        """cut"""
        # maxrth = np.amax(rth, axis=1)
        # where = np.where(maxrth < r_cap*1000)[0]
        # npart = len(where)
        x1 = xth
        y1 = y
        z1 = zth
        bx1 = bexth
        by1 = by
        bz1 = bezth
        return x1,y1,z1,bx1,by1,bz1    
    
    def SI_CGS(self,seg):
        pos_CGS = copy.deepcopy(seg.pos)*1e2
        E_CGS = copy.deepcopy(seg.E_field)*1e-4/2.9979
        B_CGS = copy.deepcopy(seg.B_field)*1e4
        return pos_CGS,E_CGS,B_CGS
    
    def add_detectors(self,segdict):
        """
        Adds radiation detectors to desired beamline segments. 
        segdict:    dictionary with segment names as keys and lists as values. 
                    First list item should be an array of shape (N,3) with N 
                    number of detector points and their 3 spatial coordinates.
                    Second list item should be an array of shape (M) with the 
                    desired spectrum frequencies.
        """
        self.radsegdict = segdict
        for sname in self.radsegdict.keys():
            self.segments[sname].add_detectors(self.radsegdict[sname])
    
    def radiation(self,parall):
        """
        Returns d^2I / dome dOme (jackson IIedition pg669) as a function of 
        radiation frequency for each segment detector point listed in self.radsegdict
        
        """
        for i,sname in enumerate(self.slist[self.istart:]):
            if sname in self.radsegdict.keys():
                print()
                print(sname+" radiation:")
                """current segment specs"""
                i += self.istart-1
                seg_curr = self.segments[self.slist[i+1]]
                te = seg_curr.t_end
                npt = seg_curr.npts
                teval = np.linspace(0, te, npt, endpoint=True)
                """extract detector points and evaluation frequencies"""
                dets,freqs = copy.deepcopy(self.radsegdict[sname])
                """turn variables in CGS"""
                pos_CGS,E_CGS,B_CGS = self.SI_CGS(seg_curr)
                if seg_curr.kind == "CBM" or seg_curr.kind == "ABP":
                    """detectors angles respect to segment bending axis"""
                    thetas = dets[:,0]
                    """detector cartesian position in thetas-rotated reference system (CGS)"""
                    dets = dets[:,1:]*100 
                    dets[:,0] += copy.deepcopy(seg_curr.kwargs["r_bend"])*100 
                    """detectors angles and radia respect to their non-rotated reference system"""
                    thetd = np.arctan2(dets[:,2],dets[:,0])
                    rd = np.sqrt(dets[:,0]**2 + dets[:,2]**2)
                    """detectors angles and position in segment's reference system"""
                    thetd += thetas
                    dets[:,0] = rd*np.cos(thetd)
                    dets[:,2] = rd*np.sin(thetd)
                else:
                    dets *= 100                 
                """calculate radiation"""
                wt = 1
                if parall==False:
                    for d,detpos in enumerate(dets):
                        ft_array = np.zeros((len(freqs),6))
                        print("detector {}".format(d))
                        for ii in range(self.beam.npart):
                            xint = pos_CGS[:,ii,:]
                            bint = copy.deepcopy(seg_curr.bet[:,ii,:])
                            gint = copy.deepcopy(seg_curr.gam[:,ii])
                            Eint = E_CGS[:,ii,:]
                            Bint = B_CGS[:,ii,:]
                            ft_array = rad_no_parall(xint,bint,gint,Eint,Bint,wt,detpos,freqs,teval,ft_array)
                        ft_out = ft_array[:,0:3] + 1j*ft_array[:,3:6]        
                        Aampl = abs(np.linalg.norm(ft_out,axis=-1))
                        U = 2*Aampl**2               
                        """update current segmets' radiation array"""
                        seg_curr.U[:,d] = U
                        
                elif parall==True:
                    ft_array = np.zeros((len(freqs),len(dets),6))
                    xint = pos_CGS
                    bint = copy.deepcopy(seg_curr.bet)
                    gint = copy.deepcopy(seg_curr.gam)
                    Eint = E_CGS
                    Bint = B_CGS
                    ft_out = rad_parall(xint,bint,gint,Eint,Bint,wt,dets,freqs,teval,ft_array)
                    Aampl = abs(np.linalg.norm(ft_out,axis=-1))
                    U = 2*Aampl**2   
                    """update current segmets' radiation array"""
                    seg_curr.U = U
                
    def plot_beam_bendplane(self):
        plt.figure()
        l = [0]
        miny = np.inf
        maxy = -np.inf
        for i,sname in enumerate(self.slist[self.istart:]):
            seg = self.segments[sname]
            rg = self.rangelist[i]
            x,y,z,bx,by,bz = seg.pos[:,:,0],seg.pos[:,:,1],seg.pos[:,:,2],seg.bet[:,:,0],seg.bet[:,:,1],seg.bet[:,:,2]
            """rectify variables for bending elements"""
            if seg.kind == "CBM" or seg.kind == "ABP":
                x,y,z,bx,by,bz = self.rectify(x,y,z,bx,by,bz,seg.kwargs["r_bend"],0)
                x *= seg.kwargs["cff"]
                bx *= seg.kwargs["cff"]
            else:
                pass
            if np.amax(x)>maxy:
                maxy = np.amax(x)
            if np.amin(x)<miny:
                miny = np.amin(x)
            plt.plot(z+sum(l),x,alpha=0.2)
            plt.scatter(z[0]+sum(l),x[0])
            l.append(seg.kwargs["l"])
        
        """plot segment labels with colors from dict in constructor"""
        npt = 10
        h = 0.1*(maxy - miny)
        upline = np.ones(npt)*(miny - h)
        downline = np.ones(npt)*(miny - 2*h)
        for i,sname in enumerate(self.slist[self.istart:]):
            """previous current and following segment specs"""
            seg = self.segments[sname]
            xax = np.linspace(sum(l[:i+1]),sum(l[:i+2]),npt)
            col = self.color_dict[seg.kind]
            if seg.kind == "CBM" or seg.kind == "ABP":
                sname = seg.kwargs["lr"]+" "+sname
            box = plt.fill_between(xax,downline,upline,color=col,alpha=0.3)
            """place segment specs inside box"""
            (x0, y0), (x1, y1) = box.get_paths()[0].get_extents().get_points()
            plt.text((x0 + x1) / 2, (y0 + y1) / 2, sname, ha='center', va='center')
        plt.legend()
            
            


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 15 10:26:09 2024

@author: Andrea
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.special import jv
import sys
from multiprocessing import Array
import os
import h5py
from scipy.stats import median_abs_deviation as MAD
import multiprocessing as mp
import ctypes as ctp
import tqdm
from scipy.special import kv

c       = 2.99792458e8*1e2                  #cm/s
me      = 9.109e-31*1e3                     #g
e       = 4.8032e-10                        #statC
h       = 6.626176e-27                      #erg*s

npn = np.newaxis

def Hz_to_keV(Hz):
    return Hz/2.41799050402417e17

def keV_to_Hz(keV):
    return keV*2.41799050402417e17

def rad_no_parall(xint,bint,gint,Eint,Bint,wt,xO,fr,t,ft_array):
    F            = -wt*e*(Eint + np.cross(bint,Bint))
    betdot       = (F - np.sum(bint*F,axis=1)[:,npn]*bint)/(c*me*wt*gint[:,npn])
    betdot       = betdot
    x            = xint
    bet          = bint
    gam          = gint
    #gamave      = np.nanmean(gam)
    """calcolo posizioni rispetto al detector e tempi ritardati"""
    Rvec         = xO - x                                                         
    R            = np.linalg.norm(Rvec,axis=1)
    n            = Rvec/R[:,npn]
    tend         = t + R/c                                #lab time when signal started at R gets to O
    """calcolo campi: A ∝ RE (jackson IIedition pg657,pg668)"""
    betdotn      = np.sum(bet*n,axis=1)
    A1           = np.sqrt(wt)*e*(n-bet)                                        /     (gam**2 * (1-betdotn)**3 * R)[:,npn]
    A2           = np.sqrt(wt)*(e/c)*(np.cross(n,np.cross((n-bet),betdot)))     /     ((1-betdotn)**3)[:,npn]
    A            = (c/4/np.pi)**0.5*(A1 + A2) 
    try:
        Aslo = (A[1:] - A[:-1])/(tend[1:,npn] - tend[:-1,npn])  
        Aslo = np.r_[Aslo , [np.array([0,0,0])]]  #field slope between consecutive points in linear interpolation
        Aint = A - Aslo*tend[:,npn]                                      #field intercept for linear interpolation
        Aint[-1] = np.array([0,0,0])
        A = np.c_[Aslo,Aint]
        lf = len(fr)
        ft = np.zeros((lf,3), dtype=complex)
        frN = fr
        lfN = len(frN)
        for i in range(lfN):
            om = 2*np.pi*frN[i]
            t1 = tend[1:][:,npn]
            t0 = tend[:-1][:,npn]
            f1 = np.exp(1j*om*t1)*(A[:-1,0:3]*(1-1j*om*t1) - 1j*om*A[:-1,3:6])
            f0 = np.exp(1j*om*t0)*(A[:-1,0:3]*(1-1j*om*t0) - 1j*om*A[:-1,3:6])
            df = (f1-f0)
            fttot = (1/np.sqrt(2*np.pi))*np.nansum(df,axis=0)/om**2
            ft[i] = fttot
        ft_array[:,0:3] += ft.real
        ft_array[:,3:6] += ft.imag
    except Exception as exc:
        print("noooo")
        pass
    return ft_array

def rad_parall(xint,bint,gint,Eint,Bint,wt,xO,fr,t,ft_array):
    pint = bint*gint[:,:,npn]
    wt = np.ones(gint.shape)*wt
    """share arrays for parallel computing"""
    rad_make_shared_array(xint, name='my_shared_array_6')           # create shared memory array from numpy array
    XfinS = rad_get_shared_array('my_shared_array_6')               # get shared memory array as numpy array
    rad_make_shared_array(pint, name='my_shared_array_6')  
    PfinS = rad_get_shared_array('my_shared_array_6')  
    rad_make_shared_array(Eint, name='my_shared_array_6')  
    EfinS = rad_get_shared_array('my_shared_array_6') 
    rad_make_shared_array(Bint, name='my_shared_array_6')  
    BfinS = rad_get_shared_array('my_shared_array_6') 
    rad_make_shared_array(wt, name='my_shared_array_6')  
    wtfinS = rad_get_shared_array('my_shared_array_6')  
    rad_make_shared_array(t, name='my_shared_array_6')
    tS = rad_get_shared_array('my_shared_array_6') 
    rad_make_shared_array(fr, name='my_shared_array_6')  
    freqsS = rad_get_shared_array('my_shared_array_6')
    rad_make_shared_array(xO, name='my_shared_array_6')  # create shared memory array from numpy array
    xOS = rad_get_shared_array('my_shared_array_6')  # get shared memory array as numpy array
    ftM = mp.Array(ctp.c_double, len(fr)*len(xO)*6) #all frequencies, 3 real parts, 3 imag parts

    items               = [(i,j) for j in range(len(xO)) for i in range(xint.shape[1])]

    """parallel ft"""               
    with mp.Pool(initializer=rad_init, initargs=(ftM,
                                                 XfinS,
                                                 PfinS,
                                                 EfinS,
                                                 BfinS,
                                                 wtfinS,
                                                 xOS,
                                                 freqsS,
                                                 tS,)) as pool:
        results         = list(tqdm.tqdm(pool.imap(rad_task, items), total=(xint.shape[1]*len(xO))))
    
    fts                 = np.frombuffer(ftM.get_obj())
    fts                 = fts.reshape(len(fr),len(xO),6)
    ft                  = fts[:,:,0:3] + 1j*fts[:,:,3:6]      
    return ft

def rad_task(item):
    """CGS units"""

    it,dtc = item

    xint = XfinS[:,it,:]
    pint = PfinS[:,it,:]
    eint = EfinS[:,it,:]
    bint = BfinS[:,it,:]
    wt = np.nanmax(wtfinS[:,it])
    xO = xOS[dtc]
    fr = freqsS
    t = tS

    pmod        = np.linalg.norm(pint,axis=1)
    gamint      = np.sqrt(1+pmod**2)
    betint      = pint/gamint[:,npn]
 
    F           = -wt*e*(eint + np.cross(betint,bint))
    betdot      = (F - np.sum(betint*F,axis=1)[:,npn]*betint)/(c*me*wt*gamint[:,npn])
    betdot      = betdot
    x           = xint
    bet         = betint
    gam         = gamint
    
    """position respect to detector and retarded times calculation"""
    Rvec         = xO - x                                                         
    R             = np.linalg.norm(Rvec,axis=1)
    n             = Rvec/R[:,npn]
    tend         = t + R/c                                #lab time when signal started at R gets to O
    
    """fields calculation: A ∝ RE (jackson IIedition pg657,pg668)"""
    betdotn        = np.sum(bet*n,axis=1)
    A1             = np.sqrt(wt)*e*(n-bet)                                        /     (gam**2 * (1-betdotn)**3 * R)[:,npn]
    A2             = np.sqrt(wt)*(e/c)*(np.cross(n,np.cross((n-bet),betdot)))     /     ((1-betdotn)**3)[:,npn]
    A               = (c/4/np.pi)**0.5*(A1 + A2) 
    try:
        Aslo = (A[1:] - A[:-1])/(tend[1:,npn] - tend[:-1,npn])  
        Aslo = np.r_[Aslo , [np.array([0,0,0])]]                        #field slope between consecutive points in linear interpolation
        Aint = A - Aslo*tend[:,npn]                                     #field intercept for linear interpolation
        Aint[-1] = np.array([0,0,0])
        addcol = np.c_[Aslo,Aint]
        tim = tend
        A = addcol
        lf = len(fr)
        ft = np.zeros((lf,3), dtype=complex)
        frN = fr
        lfN = len(frN)
        for i in range(lfN):
            om = 2*np.pi*frN[i]
            t1 = tim[1:][:,npn]
            t0 = tim[:-1][:,npn]
            f1 = np.exp(1j*om*t1)*(A[:-1,0:3]*(1-1j*om*t1) - 1j*om*A[:-1,3:6])
            f0 = np.exp(1j*om*t0)*(A[:-1,0:3]*(1-1j*om*t0) - 1j*om*A[:-1,3:6])
            df = (f1-f0)
            fttot = (1/np.sqrt(2*np.pi))*np.nansum(df,axis=0)/om**2
            ft[i] = fttot
            # ft[i] = (1/np.sqrt(2*np.pi))*np.trapz(A*np.exp(1j*om*tim[:,npn]),tim,axis=0)
        with ftM.get_lock():
            final_array = np.frombuffer(ftM.get_obj())
            final_array = final_array.reshape((lf,len(xOS),6))
            final_array[:,dtc,0:3] += ft.real
            final_array[:,dtc,3:6] += ft.imag
    except Exception as exc:
        print(exc)
        print(xint)
        print(pint)
        print(eint)
        print(bint)
        print(wt)
        sys.quit()

def theor_und_rad(om,th,ph,LU,KU,lambU,gam0):
    """analytical undulator radiation spectrum [erg s]"""

    """support functions"""
    def term(m,xs,zt,K,om,N,th,fr):
        """single term of analytical undulator radiation summation"""
        Fm = (4*m*xs/K)**2*(jv((m-1)/2,m*xs) - jv((m+1)/2,m*xs))**2
        return Fm * sinc(nu(m,om,N,fr)/2)**2

    def sinc(x):
        return np.sin(x)/x

    def nu(m,om,N,fr):
        return 2*np.pi*N*(m - om/(2*np.pi*fr))

    """analytical undulator radiation"""
    order = 30    
    l1 = lambU/2/gam0**2*(1 + KU**2/2)
    o1 = c*2*np.pi/l1
    f1 = o1/2/np.pi    
    N = LU/lambU #3#100
    zt = -KU*om*th*np.cos(ph)/(gam0*o1)
    xs = KU**2/4/(1+KU**2/2)
    cf = N**2*e**2*gam0**2/c
    tr = 0
    for i in range(order):
        io = 2*i+1
        trp = term(io,xs,zt,KU,om,N,th,f1)   
        tr +=  trp      
    return cf*tr

def theor_synch_rad(om,rho,gam,th):
    xi = om*rho*(1/gam**2 + th**2)**(3/2)/3/c
    c1 = e**2/3/np.pi**2/c
    c2 = (om*rho/c)**2
    c3 = (1/gam**2 + th**2)**2
    ks = kv(2/3,xi)**2 + th**2*kv(1/3,xi)**2/(1/gam**2 + th**2)
    return c1*c2*c3*ks

def wtstd(values, weights):
    average = np.average(values, weights=weights)
    variance = np.average((values-average)**2, weights=weights)
    return (average, np.sqrt(variance))

def rad_init(s1,s2,s3,s4,s5,s6,s7,s8,s9):
    global ftM
    global XfinS
    global PfinS
    global EfinS
    global BfinS
    global wtfinS
    global xOS
    global freqsS
    global tS
    
    ftM = s1
    XfinS = s2
    PfinS = s3
    EfinS = s4
    BfinS = s5
    wtfinS = s6
    xOS = s7
    freqsS = s8
    tS = s9

def rad_get_shared_array(name: str, shape=None):
    mp_array = globals()[name]
    np_array = np.frombuffer(mp_array.get_obj(), dtype=np.dtype(mp_array.get_obj()._type_))
    if (shape is None) and (name + '_shape' in globals().keys()):
        shape = globals()[name + '_shape']
        shape = np.frombuffer(shape.get_obj(), dtype=int)
    if shape is not None:
        np_array = np_array.reshape(shape)
    return np_array

def rad_make_shared_array(np_array: np.ndarray, name: str):
    mp_dtype = np.ctypeslib.as_ctypes(np_array.dtype.type())._type_
    mp_array = Array(typecode_or_type=mp_dtype, size_or_initializer=int(np.prod(np_array.shape)))
    globals()[name] = mp_array
    shared_np_array = rad_get_shared_array(name, shape=np_array.shape)
    shared_np_array[:] = np_array
    mp_array_shape = Array(typecode_or_type='l', size_or_initializer=len(np_array.shape))
    globals()[name + '_shape'] = mp_array_shape
    shared_np_array = rad_get_shared_array(name + '_shape')
    shared_np_array[:] = np_array.shape
    
class H5_rad():
    """CGS unit system"""
    def __init__(self,pat,dtscale,maxpart,Nangsx,Nangsy,Nfreqs,fmin,fmax,bandwidth,detdist,nep,thxlim=None,thylim=None):
        self.pat = pat                  # h5 files directory path
        self.dtscale = dtscale                    # timestep between consecutive h5 files [s]
        self.maxpart = maxpart          # maximum desired number of analized particles
        self.Nangsx = Nangsx
        self.Nangsy = Nangsy
        self.Nfreqs = Nfreqs
        self.fmax = fmax                #maximum frequency for FT [Hz]      (sets number of interpolated timesteps)
        self.bandwidth = bandwidth
        self.detdist = detdist
        self.nep = nep                  #plasma density [cm^-3]
        self.thxlim = thxlim
        self.thylim = thylim
        self.freqs = np.logspace(fmin, fmax, self.Nfreqs)  
    
    def run(self):
        """extract files, order data and perform radiation analysis"""
        self.extract_files()
        self.sort_data()
        self.div_extimate()
        self.init_detectors()
        self.share_arrays()
        self.rad()
        
    def extract_files(self):    
        pat,dt,maxpart = self.pat,self.dt,self.maxpart
        files           = []
        for file in os.listdir(pat):
            if file.endswith('.h5'):
                files.append(file)
        files           = sorted(files)    
        zpl             = []
        xpl             = []
        IDl             = []
        wtl             = []
        Xl              = []
        Pl              = []
        El              = []
        Bl              = []
        self.supparr = np.zeros(12)                                                       
    
        """preliminary ID check"""
        for num,filenam in enumerate(files):
            f           = h5py.File(pat+'/'+filenam, "r")
            fdatagroup  = f['data'][list(f['data'].keys())[0]]                       
            ele         = fdatagroup['particles/electrons']                                 
            ID          = ele['id'][()]
            IDl.append(ID)
    
            print(num)
            
        self.dt = fdatagroup.attrs["dt"]*self.dtscale
        print()
        print("timestep =",self.dt,"s")
        """final beam total charge and energy"""
        wt0             = ele['weighting'][()]                                              #weights
        PX0             = ele['momentum']['x'][()]                                          #moments [beta*gamma] of each particle inside macroparticle
        PY0             = ele['momentum']['y'][()]                                  
        PZ0             = ele['momentum']['z'][()]    
        P0              = np.sqrt(PX0**2 + PY0**2 + PZ0**2)
        charge          = np.nansum(wt0)*1.602e-19*1e12                                     #[pC]
        Eave            = np.nansum(P0*wt0*me*c**2)/np.nansum(wt0)                          #[erg]
        Espread         = np.sqrt(np.nansum((P0*me*c**2 - Eave)**2*wt0)/np.nansum(wt0))     #[erg]
        Eave            = Eave*6.24*1e11/1e6                                                #[MeV]
        Espread         = Espread*6.24*1e11/1e6                                             #[MeV] 
        lIDl            = len(IDl)
        t               = np.arange(lIDl)*dt
        self.t          = t
        IDrange         = np.sort(np.unique(np.concatenate(IDl)))                           # total extension of ID values 
        self.IDrange    = IDrange
        IDsize          = len(IDrange)
        self.supparr[0] = IDsize
        nmacro          = max(1,IDsize/maxpart)
        self.nmacro     = nmacro
        if nmacro != 1:    
            IDrange     = np.sort(np.random.choice(IDrange, size=maxpart,replace=False))
        IDsize          = len(IDrange)      
        self.supparr[1] = IDsize
        self.supparr[2] = nmacro
        self.supparr[3] = charge
        self.supparr[4] = Eave
        self.supparr[5] = Espread
    
        """temporary arrays to be filled and ordered"""
        num             += 1
        self.Xtemp      = np.zeros((num,IDsize,3))                  #[timestep,ID,coord]
        self.Ptemp      = np.zeros((num,IDsize,3))                  #[timestep,ID,coord]
        self.Etemp      = np.zeros((num,IDsize,3))                  #[timestep,ID,coord]
        self.Btemp      = np.zeros((num,IDsize,3))                  #[timestep,ID,coord]
        self.wttemp     = np.zeros((num,IDsize))                    #[timestep,ID,coord]
        self.IDtemp     = np.zeros((num,IDsize))    
        self.maxs,self.mins = 0,np.inf                              #max and min transverse rms size
        for numm,filename in enumerate(files):    
            f           = h5py.File(pat+'/'+filename, "r")
            fdatagroup  = f['data'][list(f['data'].keys())[0]]                       
            ele         = fdatagroup['particles/electrons']                          
            
            ID          = ele['id'][()]
            whID        = np.in1d(ID,IDrange)
            ID          = ID[whID]
            X0          = ele['position']['x'][()][whID]*1e-4#                               #positions     [cm]
            Y0          = ele['position']['y'][()][whID]*1e-4#                               #positions     [cm]
            Z0          = ele['position']['z'][()][whID]*1e-4#                               #positions     [cm]
            PX0         = ele['momentum']['x'][()][whID]                                     #moments       [beta*gamma]
            PY0         = ele['momentum']['y'][()][whID]                                     #moments       [beta*gamma]
            PZ0         = ele['momentum']['z'][()][whID]                                     #moments       [beta*gamma]
            EX0         = ele['fields']['E']['x'][()][whID]*1e-4/2.9979                      #Efield        [statVolt/cm]
            EY0         = ele['fields']['E']['y'][()][whID]*1e-4/2.9979
            EZ0         = ele['fields']['E']['z'][()][whID]*1e-4/2.9979
            BX0         = ele['fields']['B']['x'][()][whID]*1e4                              #Bfield        [Gauss]
            BY0         = ele['fields']['B']['y'][()][whID]*1e4
            BZ0         = ele['fields']['B']['z'][()][whID]*1e4
            wt          = ele['weighting'][()][whID]                                         #weights
            sortarr     = np.c_[ID,X0,Y0,Z0,PX0,PY0,PZ0,EX0,EY0,EZ0,BX0,BY0,BZ0,wt]
            
            sortarr     = sortarr[sortarr[:,0].argsort()]                                    #sort by ID
            
            ID          = sortarr[:,0]
            X0          = sortarr[:,1]
            Y0          = sortarr[:,2]
            Z0          = sortarr[:,3]
            PX0         = sortarr[:,4]
            PY0         = sortarr[:,5]
            PZ0         = sortarr[:,6]
            EX0         = sortarr[:,7]
            EY0         = sortarr[:,8]
            EZ0         = sortarr[:,9]
            BX0         = sortarr[:,10]
            BY0         = sortarr[:,11]
            BZ0         = sortarr[:,12]
            wt          = sortarr[:,13]
    
            lIDi        = len(ID)
    
            X           = np.stack([X0,Y0,Z0]).transpose()
            P           = np.stack([PX0,PY0,PZ0]).transpose()
            E           = np.stack([EX0,EY0,EZ0]).transpose()
            B           = np.stack([BX0,BY0,BZ0]).transpose()
                
            self.Xtemp[numm,:lIDi,:] = X
            self.Ptemp[numm,:lIDi,:] = P
            self.Etemp[numm,:lIDi,:] = E
            self.Btemp[numm,:lIDi,:] = B
            self.wttemp[numm,:lIDi] = wt
            self.IDtemp[numm,:lIDi] = ID
    
            wtl.append(list(wt))
            Xl.append(X)
            Pl.append(P)
            El.append(E)
            Bl.append(B)
            pmodu       = np.linalg.norm(P,axis=1)
            
            try:            
                    """transverse rms sizes"""
                    avex,sigx   = wtstd(X0, wt)
                    avey,sigy   = wtstd(Y0, wt)
                    sig         = (sigx + sigy)/2
                    if sig > self.maxs:
                        self.maxs   = sig
                        self.maxsx  = sigx*10    #[mm]
                        self.maxsy  = sigy*10    #[mm]
                    if sig < self.mins:
                        self.mins   = sig
                        self.minsx  = sigx*10    #[mm]
                        self.minsy  = sigy*10    #[mm]        
                    print(pmodu.shape)
                    minp        = np.nanmin(pmodu)
                    maxp        = np.nanmax(pmodu)
                    print(numm, ID.shape, "min p = ", minp, "max p = ", maxp)
            except:
                    print(numm, ID.shape)
            self.lIDl   = lIDl
            self.IDsize = IDsize
                    
    def sort_data(self):    
        lIDl            = self.lIDl
        IDsize          = self.IDsize
        self.Xfin       = np.zeros((lIDl,IDsize,3))*np.nan                  #[timestep,ID,coord]
        self.Pfin       = np.zeros((lIDl,IDsize,3))*np.nan                  #[timestep,ID,coord]
        self.Efin       = np.zeros((lIDl,IDsize,3))*np.nan                  #[timestep,ID,coord]
        self.Bfin       = np.zeros((lIDl,IDsize,3))*np.nan                  #[timestep,ID,coord]
        self.wtfin      = np.zeros((lIDl,IDsize))*np.nan                    #[timestep,ID,coord]
        
        checkarr        = np.zeros(lIDl).astype(int)                        #0th ID 
        indarr          = np.arange(lIDl)                                   #time indexes
        
        for i in range(IDsize):
            IDc         = self.IDtemp[indarr,checkarr]
            mask        = IDc - self.IDrange[i]
            mask1       = abs(np.sign(mask)).astype(int)
            mask2       = 1 - mask1                                         # 1 only where local ID equals ith ID
            indxrow     = (mask2*(indarr+1))-1                              # +1 so that this array is zero only where IDs don't match, -1 to then set missing IDs to -1
            dove        = (indxrow != -1)
            indxrow     = indxrow[dove]
            indxcol     = np.ones_like(indxrow)*i
        
            temprow     = indxrow
            tempcol     = checkarr[dove]
        
            checkarr    += mask2
        
            self.wtfin[indxrow,indxcol] = self.wttemp[temprow,tempcol]
            self.Xfin[indxrow,indxcol]  = self.Xtemp[temprow,tempcol]
            self.Pfin[indxrow,indxcol]  = self.Ptemp[temprow,tempcol]
            self.Efin[indxrow,indxcol]  = self.Etemp[temprow,tempcol]
            self.Bfin[indxrow,indxcol]  = self.Btemp[temprow,tempcol]
            print(i)
            
    def div_extimate(self):
        """radiation divergence extimation"""
        wts         = np.nanmax(self.wtfin,axis=0) #[npart]
        xs          = self.Xfin[:,:,0]
        wnn         = ~np.isnan(xs)

        pmod        = np.linalg.norm(self.Pfin,axis=2)
        gamint      = np.sqrt(1+pmod**2)
        betint      = self.Pfin/gamint[:,:,npn]
        kbs         = np.sqrt(2*np.pi*self.nep*e**2/gamint/me/c**2)

        kbs         = kbs[wnn]    
        gamint      = gamint[wnn]
        xs          = xs[wnn]
        ys          = self.Xfin[:,:,1]
        ys          = ys[wnn]
        rs          = np.sqrt(xs**2 + ys**2)
        bxs         = betint[:,:,0]
        bxs         = bxs[wnn]
        bys         = betint[:,:,1]
        bys         = bys[wnn]
        bs          = np.sqrt(bxs**2 + bys**2)
        r0s         = np.sqrt(rs**2 + bs**2/kbs**2)
        x0s         = np.sqrt(xs**2 + bxs**2/kbs**2)
        y0s         = np.sqrt(ys**2 + bys**2/kbs**2)
        r0MAD       = MAD(r0s)
        x0MAD       = MAD(x0s)
        y0MAD       = MAD(y0s)
        self.x0MAD  = x0MAD
        self.y0MAD  = y0MAD
        print()
        print("std r=",r0MAD)
        print("std x=",x0MAD)
        print("std y=",y0MAD)
        ths         = kbs
        wtg         = gamint**4*kbs**4*r0s**2
        this        = np.nansum(ths*wtg,axis=0)/np.nansum(wtg,axis=0)
        gcoeff      = np.nansum(this*wts)/np.nansum(wts)
        thxrad      = x0MAD*gcoeff
        thyrad      = y0MAD*gcoeff
        
        if self.thxlim==None:
            self.thxlim = thxrad*1.5
        if self.thylim==None:            
            self.thylim = thyrad*1.5
        print()
        print("x rad angle =",thxrad)
        print("y rad angle =",thyrad)
        print()
        
    def init_detectors(self):
        """needed attributes"""
        self.angx               = np.linspace(-self.thxlim,self.thxlim,self.Nangsx)
        self.angy               = np.linspace(-self.thylim,self.thylim,self.Nangsy)
        self.Nangs              = len(self.angx)*len(self.angy)
        AX,AY                   = np.meshgrid(self.angx,self.angy)
        Ax,Ay                   = AX.flatten(),AY.flatten()
        A                       = np.sqrt(Ax**2 + Ay**2)
        detZ                    = self.detdist*np.cos(A)
        detX                    = self.detdist*np.sin(Ax)
        detY                    = self.detdist*np.sin(Ay)
        self.FLUXangs           = np.empty((self.Nangs,self.Nfreqs))*np.nan
        self.dWdO               = np.empty(self.Nangs)*np.nan
        self.xO                      = np.stack([detX,detY,detZ]).transpose()
        
    def share_arrays(self):
        """share arrays for parallel computing"""
        rad_make_shared_array(self.Xfin, name='my_shared_array_6')           # create shared memory array from numpy array
        self.XfinS = rad_get_shared_array('my_shared_array_6')               # get shared memory array as numpy array
        rad_make_shared_array(self.Pfin, name='my_shared_array_6')  
        self.PfinS = rad_get_shared_array('my_shared_array_6')  
        rad_make_shared_array(self.Efin, name='my_shared_array_6')  
        self.EfinS = rad_get_shared_array('my_shared_array_6') 
        rad_make_shared_array(self.Bfin, name='my_shared_array_6')  
        self.BfinS = rad_get_shared_array('my_shared_array_6') 
        rad_make_shared_array(self.wtfin, name='my_shared_array_6')  
        self.wtfinS = rad_get_shared_array('my_shared_array_6')  
        rad_make_shared_array(self.t, name='my_shared_array_6')
        self.tS = rad_get_shared_array('my_shared_array_6') 
        rad_make_shared_array(self.freqs, name='my_shared_array_6')  
        self.freqsS = rad_get_shared_array('my_shared_array_6')
        rad_make_shared_array(self.xO, name='my_shared_array_6')  
        self.xOS = rad_get_shared_array('my_shared_array_6') 
        self.ftM = mp.Array(ctp.c_double, len(self.freqs)*len(self.xO)*6)
        
    def rad(self):               
        self.supparr[6]         = self.Nangsx
        self.supparr[7]         = self.Nangsy
        self.supparr[8]         = self.minsx
        self.supparr[9]         = self.minsy
        self.supparr[10]        = self.maxsx
        self.supparr[11]        = self.maxsy
        
        items               = [(i,j) for j in range(len(self.xO)) for i in range(self.Xfin.shape[1])]

        """parallel ft"""               
        with mp.Pool(initializer=rad_init, initargs=(self.ftM,
                                                     self.XfinS,
                                                     self.PfinS,
                                                     self.EfinS,
                                                     self.BfinS,
                                                     self.wtfinS,
                                                     self.xOS,
                                                     self.freqsS,
                                                     self.tS,)) as pool:
            results         = list(tqdm.tqdm(pool.imap(rad_task, items), total=(self.Xfin.shape[1]*len(self.xO))))
        
        fts                 = np.frombuffer(self.ftM.get_obj())
        fts                 = fts.reshape(len(self.freqs),len(self.xO),6)
        ft                  = fts[:,:,0:3] + 1j*fts[:,:,3:6]      
        
        Aampl = abs(np.linalg.norm(ft,axis=-1))
        U = 2*Aampl**2                                                      #d^I/domedOme (jackson IIedition pg669)
    
        """photon number per solid angle"""                               
        self.dWdOmefreq     = np.trapz(U,2*np.pi*self.freqs,axis=0)         #total energy per unit solid angle, integrated from frequency spectrum
        self.FLUXangs       = U*self.nmacro
        self.dWdO           = self.dWdOmefreq*self.nmacro
        
    def wtstd(self,values, weights):
        average = np.average(values, weights=weights)
        variance = np.average((values-average)**2, weights=weights)
        return (average, np.sqrt(variance))
    


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 10:17:35 2024

@author: Andrea
"""
import numpy as np
from scipy.constants import c,m_e,e,mu_0,epsilon_0
from .beam_utils import beam
import radyno.utils.fields_utils as fu

class segment():
    def __init__(self,device,name,**kwargs):
        self.field = fu.field_dict[device]
        self.kind = device
        self.name = name
        self.kwargs = self.check_kwargs(device,kwargs)
        
        """simulation times and evaluation points"""
        self.t_end = self.kwargs["l"]/c
        
        """utils"""
    
    def add_timesteps(self,dt):
        """beam dynamics simulation timestep value and number"""
        self.dt = dt
        self.npts = int(self.t_end/dt)
        
    def init_arrays(self,npart):
        """
        init empty arrays for segment beam variables storage: note that segment
        should first have a self.npts attribute, so self.add_timesteps should 
        first be run
        """
        # assert hasattr(self.__class__, 'self.npts')
        self.pos = np.zeros((self.npts,npart,3))
        self.sigpos = np.zeros((self.npts,3))
        self.mupos = np.zeros((self.npts,3))
        self.bet = np.zeros((self.npts,npart,3))
        self.gam = np.zeros((self.npts,npart))
        self.E_field = np.zeros((self.npts,npart,3))
        self.B_field = np.zeros((self.npts,npart,3))
    
    def check_kwargs(self,device,kwargs):
        d = {"cff":1}
        kwargs.update(d)
        
        if device=="drift":
            assert "l" in kwargs
            r_bend = 0
            d = {"r_bend":r_bend}
            kwargs.update(d)
        
        elif device=="undulator":
            assert "K" in kwargs and "l_U" in kwargs and "l" in kwargs and "gamma" in kwargs
            B0 = kwargs['K']*2*np.pi*m_e*c/e/kwargs['l_U']              # magnetic field max amplitude [T]
            k_U = 2*np.pi/kwargs['l_U']                                 # undulator wavevector [m^-1]
            l_1 = kwargs['l_U']/2/kwargs["gamma"]**2*(1 + kwargs["K"]**2/2)        # first harmonic wavelength [m^-1]
            r_bend = 0
            d = {"B0":B0,
                 "k_U":k_U,
                 "l_1":l_1,
                 "r_bend":r_bend}
            kwargs.update(d)
        
        elif device=="ion channel II":
            assert "l_b" in kwargs and "K" in kwargs and "l" in kwargs and "gamma" in kwargs
            
            gamma = kwargs["gamma"] 
            K = kwargs["K"]
            
            C = np.arctanh(K/gamma)**2
           
            dg = gamma*C/(2+0*np.pi/21*C)
            g0 = gamma
            gamma = gamma + dg/2
            l_b_set = kwargs["l_b"]/(1 - 7/15*dg/gamma)/(1 + np.pi/24*dg/gamma)
            
            n_p = 8*np.pi**2*epsilon_0*m_e*c**2*gamma/e**2/l_b_set**2
            
            omega_p = np.sqrt(n_p*e**2/m_e/epsilon_0)
            k_b = 2*np.pi/kwargs["l_b"]
            
            r_off = np.arctanh(kwargs["K"]/g0)/k_b
            
            r_bend = 0
            d = {"omega_p":omega_p,
                 "k_b":k_b,
                 "r_off":r_off,
                 "r_bend":r_bend,
                 "n_p":n_p}
            kwargs.update(d)
            
            omega_b = omega_p/np.sqrt(2*gamma)
            print(2*np.pi/omega_b)
            
        elif device=="ion channel I":
            assert "l_b" in kwargs and "K" in kwargs and "l" in kwargs and "gamma" in kwargs
            n_p = 8*np.pi**2*epsilon_0*m_e*c**2*(kwargs["gamma"])/e**2/kwargs["l_b"]**2/(1 + 4.5/4/100)
            omega_p = np.sqrt(n_p*e**2/m_e/epsilon_0)
            k_b = 2*np.pi/kwargs["l_b"]
            r_off = kwargs["K"]/kwargs["gamma"]/k_b
            r_bend = 0
            d = {"omega_p":omega_p,
                  "k_b":k_b,
                  "r_off":r_off,
                  "r_bend":r_bend,
                  "n_p":n_p}
            kwargs.update(d)
        
        elif device=="CBM":
            assert "B0" in kwargs or "r_bend" in kwargs
            assert "lr" in kwargs and "th_bend" in kwargs and "gamma" in kwargs
            aveb = np.sqrt(1-1/kwargs["gamma"]**2)
            if kwargs["lr"] == "L":
                kwargs["cff"] = -1
            if "B0" in kwargs:
                r_bend = aveb*kwargs["gamma"]*m_e*c/kwargs["B0"]/e
                l = r_bend*kwargs["th_bend"]
                d = {"l":l,
                     "r_bend":r_bend}
            elif "r_bend" in kwargs:
                B0 = aveb*kwargs["gamma"]*m_e*c/kwargs["r_bend"]/e
                l = kwargs["r_bend"]*kwargs["th_bend"]
                d = {"l":l,
                     "B0":B0}
            kwargs.update(d)

        elif device=="ABP":
            """change J with I"""
            assert "r_c" in kwargs and "rho_c" in kwargs and "J" in kwargs and "lr" in kwargs and "th_bend" in kwargs and "gamma" in kwargs
            B1 = mu_0*kwargs["J"]/2                                # field slope [T/m]
            I = kwargs["J"]*np.pi*kwargs["r_c"]**2                           # discharge current [A]
            if kwargs["lr"] == "L":
                kwargs["cff"] = -1
            aveb = np.sqrt(1-1/kwargs["gamma"]**2)
            r_bend = kwargs["rho_c"]/2*(1+np.sqrt(1 + aveb*kwargs["gamma"]*m_e*c*8*np.pi*kwargs["r_c"]**2/(mu_0*e*kwargs["rho_c"]**2*I)))
            l = self.r_bend*self.th_bend
            d = {"B1":B1,
                 "I":I,
                 "l":l,
                 "r_bend":r_bend}
            kwargs.update(d)
        
        return kwargs
    
    def add_detectors(self,specs):
        """
        Init evaluation frequencies array and empty array for differential 
        intensity from each detector
        """
        dets,freqs = specs
        self.freqs = freqs
        self.U = np.zeros((len(freqs),len(dets)))
    
    def match(self):
        """overwrites original segment kwargs and matches parameters to given beam"""



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 15 10:27:04 2024

@author: Andrea
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.constants import c,m_e,e,mu_0,epsilon_0
from scipy.integrate import solve_ivp
from .rad_utils import rad_parall,rad_no_parall

class beam():
    """default: gaussian beam"""
    def __init__(self,sig,mu,aveg,sigdg,epsx,epsy,npart):
        """default: gaussian beam"""
        self.mux = mu[0]        # x rms size [m]
        self.muy = mu[1]        # y rms size [m]
        self.muz = mu[2]        # z rms size [m]
        self.sigx = sig[0]      # x rms size [m]
        self.sigy = sig[1]      # y rms size [m]
        self.sigz = sig[2]      # z rms size [m]
        self.aveg = aveg        # average Lorentz factor
        self.sigdg = sigdg      # rms relative energy spread dgamma/gamma
        self.epsx = epsx        # x normalized emittance [mm mrad]
        self.epsy = epsy        # y normalized emittance [mm mrad] 
        self.npart = npart      # number of beam particles

        """calculated variables"""
        self.sigg = aveg*sigdg                                              # rms energy spread dgamma
        self.epsxr = np.sqrt(epsx**2/aveg**2/(1 + self.sigg**2/aveg**2))    # x rms emittance [mm mrad]
        self.epsyr = np.sqrt(epsy**2/aveg**2/(1 + self.sigg**2/aveg**2))    # y rms emittance [mm mrad]
    
        """gammas and positions"""
        self.gamma = np.random.normal(aveg,self.sigg,(npart))
        self.pos0 = np.random.normal(mu, sig, (npart, 3))          # particle initial positions [m]
        
        """angles: calculated throug normalized emittance expression with gamma spread WITHOUT trace space correlation"""
        try:
            self.sigthx = epsx*1e-6/self.sigx/aveg/np.sqrt(self.sigg**2/aveg**2+1)
        except:
            self.sigthx = 0
        try:
            self.sigthy = epsy*1e-6/self.sigy/aveg/np.sqrt(self.sigg**2/aveg**2+1)
        except:
            self.sigthy = 0
        self.thx = np.random.normal(0, self.sigthx, (npart))
        self.thy = np.random.normal(0, self.sigthy, (npart))
        self.th = np.sqrt(self.thx**2 + self.thy**2)
        self.ph = np.arctan2(self.thy, self.thx)
        
        """normalized velocities"""
        beta = np.sqrt(1-1/self.gamma**2)
        bx0 = beta*np.sin(self.th)*np.cos(self.ph)
        by0 = beta*np.sin(self.th)*np.sin(self.ph)
        bz0 = beta*np.cos(self.th)
        self.bet0 = np.zeros_like(self.pos0)
        self.bet0[:,0] = bx0
        self.bet0[:,1] = by0
        self.bet0[:,2] = bz0
        
    def plot(self,var,c0,c1):
        plt.figure()
        if var=="pos":
            xp = self.pos[:,:,c0]
            yp = self.pos[:,:,c1]
        if var=="sigpos":
            xp = self.mupos[:,c0]
            yp = self.sigpos[:,c1]
        if var=="bet":
            xp = self.bet[:,:,c0]
            yp = self.bet[:,:,c1]
        plt.plot(xp,yp,alpha=0.5)
        plt.scatter(xp[0],yp[0])


# -*- coding: utf-8 -*-
"""
Created on Mon Apr  1 16:15:42 2024

@author: Andrea
"""
from .beamline_utils import beamline
from .beam_utils import beam
from .segment_utils import segment
from .fields_utils import defdict,no_field,undulator_field,plasma_field,CBM_field,ABP_field,heaviside_close,spatial_field_ramp
from .rad_utils import H5_rad,Hz_to_keV,rad_no_parall,rad_parall,theor_und_rad,theor_synch_rad,wtstd,rad_init,rad_task,rad_make_shared_array,rad_get_shared_array




# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 09:54:18 2024

@author: Andrea
"""
import numpy as np
from scipy.constants import c,m_e,e,mu_0,epsilon_0

"""fields for several beamline elements and spatial ramp"""

def defdict():
    field_dict = {"drift":         no_field,
                  "undulator":     undulator_field,
                  "ion channel":   plasma_field,
                  "CBM":           CBM_field,
                  "ABP":           ABP_field}
    return field_dict

def no_field(x,y,z,**kwargs):
    zero = np.zeros_like(x)
    return zero,zero,zero,zero,zero,zero

def undulator_field(x,y,z,**kwargs):
    Ex = np.zeros_like(x)
    Ey = np.zeros_like(x)
    Ez = np.zeros_like(x)
    Bx = np.zeros_like(x)
    By = kwargs["B0"]*np.cos(kwargs["k_U"]*z)
    # By = kwargs["B0"]*np.cos(kwargs["k_U"]*(z))
    Bz = np.zeros_like(x)
    return Ex,Ey,Ez,Bx,By,Bz

def plasma_field(x,y,z,**kwargs):
    r = np.sqrt((x - kwargs["r_off"])**2 + y**2)
    theta = np.arctan2(y,(x - kwargs["r_off"]))
    Er = kwargs["n_p"]*e*r/2/epsilon_0
    Ex = Er*np.cos(theta)
    Ey = Er*np.sin(theta)
    Ez = np.zeros_like(x)
    Bx = np.zeros_like(x)
    By = np.zeros_like(x)
    Bz = np.zeros_like(x)
    return Ex,Ey,Ez,Bx,By,Bz    

def CBM_field(x,y,z,**kwargs):
    Ex = np.zeros_like(x)
    Ey = np.zeros_like(x)
    Ez = np.zeros_like(x)
    Bx = np.zeros_like(x)
    By = -kwargs["B0"]*np.ones_like(x)#*kwargs["cff"]
    Bz = np.zeros_like(x)
    return Ex,Ey,Ez,Bx,By,Bz

def ABP_field(x,y,z,**kwargs):
    r = np.sqrt(x**2 + z**2)
    dr = r-kwargs["rho_c"]
    drax = np.sqrt(dr**2 + y**2)            # distance from design trajectory [m]
    the = np.arctan2(y, dr)                 # azimuthal angle around capillary axis[rad]
    phi = np.arctan2(z, x)                  # angle around capillary bending axis[rad]
    Bmod = kwargs["B1"]*drax
    By = Bmod*np.cos(the)*kwargs["cff"] 
    Br = Bmod*np.sin(the)*(-kwargs["cff"])
    Bx = Br*np.cos(phi)
    Bz = Br*np.sin(phi)
    Ex = np.zeros_like(x)
    Ey = np.zeros_like(x)
    Ez = np.zeros_like(x)   
    return Ex,Ey,Ez,Bx,By,Bz

def heaviside_close(x1, x2):
    closeCheck = np.isclose(x1, np.zeros_like(x1))
    heavisideBare = np.heaviside(x1, 0.0)
    zeroVal = np.where(closeCheck, x2, 0.0)-np.where(closeCheck, heavisideBare, np.zeros_like(heavisideBare))
    result = heavisideBare+zeroVal
    return result

def spatial_field_ramp(z,z0,updown):
    """
    Fermi-Dirac function for spatial field linking.
    
    z: longitudinal coordinate [m]
    z0: field linking point [m]
    updown: 1 for upramp, -1 for downramp
    """
    
    """linking stiffnes"""
    # a = 1e-6
    # return 1/(1 + np.exp(-updown*(z - z0)/a))
    return np.heaviside(updown*(z-z0), (1+updown)/2)
    # return heaviside_close(updown*z,z0)

"""leave as last row"""
field_dict = defdict()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 24 22:33:21 2024

@author: Andrea
"""

from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.8'
DESCRIPTION = 'Particles tracking and radiation spectrum evaluation'

# Setting up
setup(
    name="radyno",
    version=VERSION,
    author="exborgg (Andrea Frazzitta)",
    author_email="<andrea.frazzitta@uniroma1.it>",
    description=DESCRIPTION,
    packages=find_packages(),
    #install_requires=['opencv-python', 'pyautogui', 'pyaudio'],
    keywords=['python', 'beam', 'radiation', 'spectrum', 'mpi', 'parallel'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)